package Autofont::VPL;
use FileHandle;

sub new (;$$) {
    my $class = shift;
    my $self = { };
    bless $self, $class;
    if (@_) {
	$self->read(@_);
    } else {
	$self;
    }
}

sub _to_unbalanced_right_paren ($) {
    my($t) = @_;
    my($l, $r, $depth) = (index($t, '('), index($t, ')'), 0);
    die if $r < 0;
    while ($depth || ($l >= 0 && $l < $r)) {
	if ($r < $l || $l < 0) {
	    $depth--;
	    $r = index($t, ')', $r + 1);
	    die if $depth && $r < 0;
	} elsif ($l < $r) {
	    $depth++;
	    $l = index($t, '(', $l + 1);
	}
    }
    (substr($t, 0, $r), substr($t, $r + 1));
}

sub read ($$;$) {
    my($vpl, $fh, $fontname) = @_;

    # read VPL file
    my($fh_provided) = (ref($fh));
    if (!$fh_provided) {
	my($fn) = $fh;
	$fh = new FileHandle $fn, "r";
	return undef if !defined($fh);
    }
    local($/) = undef;
    my($v) = <$fh>;
    $fh->close if !$fh_provided;

    # canonicalize it
    # remove spaces after open parens, uppercase first words
    $v =~ s/\(\s*(\S+)/\(\U$1\E/sg;

    # remove comments; assume no comments have doubly-nested parentheses
    $v =~ s/\(COMMENT\s+([^()]|\([^()]*\))*\)//sg;
    $v =~ s/\(COMMENT\)//sg;
    die if $v =~ /\(COMMENT/;

    # get rid of 'C' and 'O' after LABEL, KRN, LIG, CHARACTER
    $v =~ s/\((LABEL|KRN|LIG|CHARACTER|MAPFONT)\s+C\s+(\S)/"\($1 D " . ord($2)/sge;
    $v =~ s/\((LABEL|KRN|LIG|CHARACTER|MAPFONT)\s+O\s+(\d+)/"\($1 D " . oct($2)/sge;

    # same for second argument to LIG
    $v =~ s/\((LIG\s+D\s+\d+)\s+C\s+(\S)/"\($1 D " . ord($2)/sge;
    $v =~ s/\((LIG\s+D\s+\d+)\s+O\s+(\d+)/"\($1 D " . oct($2)/sge;

    # get rid of 'C', 'O', and 'D' after CHARWD, CHARDP, CHARHT, CHARIC
    $v =~ s/\(CHAR(WD|DP|HT|IC)\s+C\s+(\S)/"\(CHAR$1 R " . ord($2)/sge;
    $v =~ s/\(CHAR(WD|DP|HT|IC)\s+O\s+(\d+)/"\(CHAR$1 R " . oct($2)/sge;
    $v =~ s/\(CHAR(WD|DP|HT|IC)\s+D\s+/(CHAR$1 R /sg;

    # now parse it
    my($i, $b);
    $v =~ s/\A\s+//s;
    while ($v ne '') {
	die if $v !~ /\A\(/;
	if ($v =~ /\A\(CHARACTER\s+D\s+(\d+)\s+(.*)\Z/s) {
	    die if $vpl->{'CHARACTER'}->[$1];
	    ($vpl->{'CHARACTER'}->[$1], $v) = _to_unbalanced_right_paren($2);
	} elsif ($v =~ /\A\(MAPFONT\s+D\s+(\d+)\s+(.*)\Z/s) {
	    die if $vpl->{'MAPFONT'}->[$1];
	    ($vpl->{'MAPFONT'}->[$1], $v) = _to_unbalanced_right_paren($2);
	} elsif ($v =~ /\A\((\S+)\s+(.*)\Z/s) {
	    die if $vpl->{$1};
	    ($vpl->{$1}, $v) = _to_unbalanced_right_paren($2);
	} else {
	    die ">$v";
	}
	$v =~ s/\A\s+//s;
    }

    if ($fontname) {
	$vpl->{'FONTNAME'} = $fontname;
	$vpl->{'MAPFONT'}->[0] = "(FONTNAME $fontname)\n"
	    if !exists($vpl->{'MAPFONT'});
    }

    $vpl->{'LIGTABLE'} = "" if !defined($vpl->{'LIGTABLE'});

    $vpl;
}


my(%output_order) = ('VTITLE' => -1, 'MAPFONT' => 1, 'LIGTABLE' => 2, 'CHARACTER' => 3);

sub write ($$) {
    my($vpl, $fh) = @_;

    # read VPL file
    my($fh_provided) = (ref($fh));
    if (!$fh_provided) {
	my($fn) = $fh;
	$fh = new FileHandle $fn, "w";
	return undef if !defined($fh);
    }

    foreach $k (sort { ($output_order{$a} || 0) <=> ($output_order{$b} || 0) } keys %$vpl) {
	next if $k eq 'FONTNAME' || $k eq 'encoding';
	$v = $vpl->{$k};
	if (ref($v)) {
	    for ($i = 0; $i < @$v; $i++) {
		print $fh "($k D $i\n$v->[$i])\n"
		    if defined $v->[$i];
	    }
	} else {
	    print $fh "($k $v)\n";
	}
    }
    
    $fh->close if !$fh_provided;
    1;
}


sub encoding ($) {
    my($vpl) = @_;
    $vpl->{'encoding'};
}

sub set_encoding ($$) {
    my($vpl, $enc) = @_;
    $vpl->{'encoding'} = $enc;
}


sub _skip_ligtable ($;$) {
    my($lt, $want_ligtable) = @_;
    my($orig_lt) = $lt;
    while ($lt) {
	if ($lt =~ /\A\s*\(STOP\s*\)\s*(.*)\Z/s) {
	    $lt = $1;
	    last;
	} elsif (!$want_ligtable && $lt =~ /\A\s*(\(LABEL\s+.*)\Z/s) {
	    $lt = $1;
	    last;
	} elsif ($lt =~ /\A\s*\(.*?\)\s*(.*)\Z/s) {
	    $lt = $1;
	} else {
	    die;
	}
    }
    ($want_ligtable ? substr($orig_lt, 0, length($orig_lt) - length($lt)) : $lt);
}

sub resolve_char ($@) {
    my($vpl) = shift @_;
    my(@r, $i, $out);
    
    foreach my $c (@_) {
	my(@tries) = (ref($c) ? @$c : ($c));
	for ($i = 0, $out = undef; !defined($out) && $i < @tries; $i++) {
	    if (defined($tries[$i]) && $tries[$i] =~ m|^/(.*)$|) {
		$out = $vpl->encoding()->map($1);
	    } else {
		$out = $tries[$i];
	    }
	}
	push @r, $out;
    }

    if (wantarray || @_ != 1) {
	@r;
    } else {
	$r[0];
    }
}

sub substitute_chars ($$@) {
    my($dst_vpl, $src_vpl, @chars) = @_;

    # add a MAPFONT entry
    my($fontnum) = scalar(@{$dst_vpl->{'MAPFONT'}});
    die if !$src_vpl->{'FONTNAME'};
    $dst_vpl->{'MAPFONT'}->[$fontnum] = "(FONTNAME " . $src_vpl->{'FONTNAME'} . ")";

    # map @chars onto two arrays
    my(@src_chars, @dst_chars, $x, $y, $i);
    foreach $x (@chars) {
	if (ref($x)) {
	    push @dst_chars, $x->[0];
	    push @src_chars, $x->[1];
	} else {
	    push @dst_chars, $x;
	    push @src_chars, $x;
	}
    }
    @src_chars = $src_vpl->resolve_char(@src_chars);
    @dst_chars = $dst_vpl->resolve_char(@dst_chars);
    
    # take @src_chars in $src_vpl and stick them into $dst_vpl
    my(@char_map);
    foreach my $i (0 .. $#src_chars) {
	my($s, $d) = ($src_chars[$i], $dst_chars[$i]);
	next if !defined($s) || !defined($d);
	$char_map[$s] = $d;
	my($sc) = $src_vpl->{'CHARACTER'}->[$s];
	die "no character $s in $src_vpl->{'FONTNAME'}" if !$sc;
	my($dc) = "(MAP (SELECTFONT D $fontnum) (SETCHAR D $s))\n";
	$dc .= "(CHARWD $1)\n" if $sc =~ /\(CHARWD\s+(.*?)\)/s;
	$dc .= "(CHARHT $1)\n" if $sc =~ /\(CHARHT\s+(.*?)\)/s;
	$dc .= "(CHARDP $1)\n" if $sc =~ /\(CHARDP\s+(.*?)\)/s;
	$dc .= "(CHARIC $1)\n" if $sc =~ /\(CHARIC\s+(.*?)\)/s;
	$dst_vpl->{'CHARACTER'}->[$d] = $dc;
    }
    
    # kill all @dst_chars kern and ligature programs in $dst_vpl
    my($slig, $dlig) = ($src_vpl->{'LIGTABLE'}, $dst_vpl->{'LIGTABLE'});
    foreach my $d (@dst_chars) {
	next if !defined($d);
	$dlig =~ s/\(LIG\s+D\s+$d\s+(.*?)\)\s*//sg;
	$dlig =~ s/\(LIG\s+D\s+\d+\s+D\s+$d\s*\)\s*//sg;
	$dlig =~ s/\(KRN\s+D\s+$d\s+(.*?)\)\s*//sg;
	if ($dlig =~ /\A\(LABEL\s+D\s+$d\s*\)\s*(.*)\Z/s) {
	    $dlig = _skip_ligtable($1);
	} elsif ($dlig =~ /\A(.*\(STOP\)\s*)\(LABEL\s+D\s+$d\s*\)\s*(.*)\Z/s) {
	    $dlig = $1 . _skip_ligtable($2);
	} elsif ($dlig =~ /\A(.*)\(LABEL\s+D\s+$d\s*\)\s*(.*)\Z/s) {
	    $dlig = $1 . $2;
	}
    }
    $dst_vpl->{'LIGTABLE'} = $dlig;

    # add @src_chars ligature programs from $src_vpl
    foreach my $i (0 .. $#src_chars) {
	my($s, $d) = ($src_chars[$i], $dst_chars[$i]);
	next if !defined($s) || !defined($d);
	next if $slig !~ /\(LABEL\s+D\s+$s\s*\)\s*(.*)\Z/s;
	my($sl) = _skip_ligtable($1, 1);
	my($dl) = '';
	foreach $y (split(/\(/, $sl)) {
	    if ($y =~ /\A(LABEL|STOP|\s*\Z)/s) {
		# nada
	    } elsif ($y =~ /\ALIG\s+D\s+(\d+)\s+D\s+(\d+)\s*\)/s) {
		if (defined($char_map[$1]) && defined($char_map[$2])) {
		    $dl .= "(LIG D " . $char_map[$1] . " D " . $char_map[$2] . ")\n";
		}
	    } elsif ($y =~ /\AKRN\s+D\s+(\d+)\s+(.*?)\)/s) {
		if (defined($char_map[$1])) {
		    $dl .= "(KRN D " . $char_map[$1] . " " . $2 . ")\n";
		}
	    } else {
		die;
	    }
	}
	if ($dl) {
	    $dst_vpl->{'LIGTABLE'} .= "(LABEL D $d)\n$dl\n(STOP)\n";
	}
    }

    # done!
    1;
}

sub add_lig ($$$$) {
    my($vpl) = shift @_;
    my($c1, $c2, $d) = $vpl->resolve_char(@_);
    return if !defined($c1) || !defined($c2) || !defined($d);
    my($lig) = $vpl->{'LIGTABLE'};

    # add a LABEL for $c1 if necessary
    if ($lig !~ /\(LABEL\s+D\s+$c1\s*\)/) {
	$lig .= "(LABEL D $c1)\n(STOP)\n";
    }

    # extract parts of the program
    # XXX assume no overlap in programs
    my($before, $l, $after);
    ($before, $l) = ($lig =~ /\A(.*)\(LABEL\s+D\s+$c1\s*\)\s*(.*)\Z/s);
    ($l, $after) = ($l =~ /\A(.*?\(STOP\s*\)\s*)(.*)\Z/s);

    # remove old kerns and/or ligatures
    $l =~ s/\(LIG\s*D\s*$c2\s.*?\)\s*//g;
    $l =~ s/\(KRN\s*D\s*$c2\s.*?\)\s*//g;

    # add new ligature
    $l = "(LIG D $c2 D $d)\n$l";

    # copy it back
    $vpl->{'LIGTABLE'} = "$before(LABEL D $c1)\n$l$after";
}

sub remove_lig ($$$) {
    my($vpl) = shift @_;
    my($c1, $c2) = $vpl->resolve_char(@_);
    return if !defined($c1) || !defined($c2);
    my($lig) = $vpl->{'LIGTABLE'};

    # skip out if no program for $c1
    return if (!defined($lig) || $lig !~ /\(LABEL\s+D\s+$c1\s*\)/);

    # extract parts of the program
    # XXX assume no overlap in programs
    my($before, $l, $after);
    ($before, $l) = ($lig =~ /\A(.*)\(LABEL\s+D\s+$c1\s*\)\s*(.*)\Z/s);
    ($l, $after) = ($l =~ /\A(.*?\(STOP\s*\)\s*)(.*)\Z/s);

    # remove ligature
    $l =~ s/\(LIG\s*D\s*$c2\s.*?\)\s*//g;

    # copy it back
    $vpl->{'LIGTABLE'} = "$before(LABEL D $c1)\n$l$after";
}

sub copy_right_kerns ($$$) {
    my($vpl) = shift @_;
    my($d, $c) = $vpl->resolve_char(@_);
    return if !defined($d) || !defined($c);
    my($lig) = $vpl->{'LIGTABLE'};

    # ignore if $c has no right kerns
    return if $lig !~ /\(LABEL\s+D\s+$c\s*\)(.*)\Z/s;
    my($c_prog) = _skip_ligtable($1, 1);

    # add a LABEL for $d if necessary
    if ($lig !~ /\(LABEL\s+D\s+$d\s*\)/s) {
	$lig .= "(LABEL D $d)\n(STOP)\n";
    }

    # extract parts of the program
    # XXX assume no overlap in programs
    my($before, $l, $after);
    ($before, $l) = ($lig =~ /\A(.*)\(LABEL\s+D\s+$d\s*\)\s*(.*)\Z/s);
    ($l, $after) = ($l =~ /\A(.*?\(STOP\s*\)\s*)(.*)\Z/s);

    # add new kerns
    while ($c_prog =~ /\(KRN\s*D\s*(\d+)\s*(.*?)\)/sg) {
	my($k, $amt) = ($1, $2);
	next if $l =~ /\((LIG|KRN)\s*D\s*$k\s/s;
	$l = "(KRN D $k $amt)\n$l";
    }

    # copy it back
    $vpl->{'LIGTABLE'} = "$before(LABEL D $d)\n$l$after";
}

sub copy_left_kerns ($$$) {
    my($vpl) = shift @_;
    my($d, $c) = $vpl->resolve_char(@_);
    return if !defined($d) || !defined($c);
    my($lig) = $vpl->{'LIGTABLE'};

    # copy into a new lig string
    my($new_lig) = '';
    while ($lig =~ /\A(.*?)\((LIG|KRN)\s*D\s*$c\s+(.*?)\)\s*(.*)\Z/s) {
	$new_lig .= "$1($2 D $c $3)\n($2 D $d $3)\n";
	$lig = $4;
    }
    $new_lig .= $lig;

    # copy it back
    $vpl->{'LIGTABLE'} = $new_lig;
}

sub each_char ($&) {
    my($vpl, $func) = @_;
    my($i);
    for ($i = 0; $i < 256; $i++) {
	if ($vpl->{'CHARACTER'}->[$i]) {
	    local($_) = $vpl->{'CHARACTER'}->[$i];
	    &$func();
	    $vpl->{'CHARACTER'}->[$i] = $_;
	}
    }
}

1;
